
#include "bma456_node.h"

// Define the previously declared static variables
SPISettings BMA456::_spiSettings = SPISettings();
uint8_t BMA456::_pin_cs          = -1;

BMA456::BMA456(){
    // init();
};

/**
 * @brief
 */
int8_t BMA456::init(uint8_t clk, uint8_t miso, uint8_t mosi, uint8_t ss, uint8_t int1, uint8_t int2)
{
    int8_t rslt = BMA4_OK;

    spi_init(clk, miso, mosi, ss);

    bma.intf           = BMA4_SPI_INTF;
    bma.bus_read       = spi_read;
    bma.bus_write      = spi_write;
    bma.intf_ptr       = &dev_addr;  //  Assign device address to interface pointer
    bma.delay_us       = delay_us;
    bma.read_write_len = 8;
    bma.resolution     = 16;
    bma.feature_len    = BMA456H_FEATURE_SIZE;
    bma.variant        = BMA45X_VARIANT;

    // Verify chip ID
    rslt |= bma456h_init(&bma);
    if (rslt != BMA4_OK)
    {
        bma4_error_codes_print_result("Initialisation", rslt);
        return rslt;
    }
    else
        log_i("Initialisation success");

    bma4_soft_reset(&bma);
    vTaskDelay(200);
    getID();

    // Configuration
    rslt |= bma456h_write_config_file(&bma);
    if (rslt != BMA4_OK)
    {
        // log_e("Load configuration fail\r\n");
        bma4_error_codes_print_result("Load configuration", rslt);
        return rslt;
    }
    else
    {
        log_i("Load configuration successful\r\n");
    }
    vTaskDelay(5);

    // Accelerometer ////////////////////////////

    rslt |= bma4_set_accel_enable(BMA4_ENABLE, &bma);
    bma4_error_codes_print_result("bma4_set_accel_enable status", rslt);
    vTaskDelay(5);

    /* Accelerometer configuration settings */
    accel_conf.odr       = BMA4_OUTPUT_DATA_RATE_200HZ;
    accel_conf.range     = BMA4_ACCEL_RANGE_4G;
    accel_conf.bandwidth = BMA4_ACCEL_NORMAL_AVG4;
    accel_conf.perf_mode = BMA4_CONTINUOUS_MODE;
    rslt |= bma4_set_accel_config(&accel_conf, &bma);
    bma4_error_codes_print_result("bma4_set_accel_enable status", rslt);

    // rslt |= bma456_map_interrupt(BMA4_INTR1_MAP, BMA4_DATA_RDY_INT, BMA4_ENABLE, &bma);
    // bma4_error_codes_print_result("bma456_map_interrupt status", rslt);

    tap_init();

    // FIFO
    /* Mapping data ready interrupt with interrupt pin 1 to get interrupt status once getting new accel data */
    // Interrupts for fifo
    // rslt |= bma456_map_interrupt(int_line_fifo, int_map_fifo, BMA4_ENABLE, &bma);
    // bma4_error_codes_print_result("bma456_map_interrupt status", rslt);

    // uint8_t* feature_config;
    // rslt = bma4_read_regs(BMA4_FEATURE_CONFIG_ADDR, feature_config, BMA456_FEATURE_SIZE, &bma);

    return rslt;
}

void BMA456::accTest()
{
    uint16_t rslt = BMA4_OK;
    Serial.printf("Ax[m/s2], Ay[m/s2], Az[m/s2]\r\n");
    n_data = ACCEL_SAMPLE_COUNT;

    while (1)  // Attempt n_data times
    {
        /* Read interrupt status */
        rslt = bma456h_read_int_status(&int_status, &bma);
        bma4_error_codes_print_result("bma456_read_int_status", rslt);

        /* Filtering only the accel data ready interrupt */
        if ((rslt == BMA4_OK) && (int_status & BMA4_ACCEL_DATA_RDY_INT))
        {
            /* Read the accel x, y, z data */
            rslt = bma4_read_accel_xyz(&sens_data, &bma);
            bma4_error_codes_print_result("bma4_read_accel_xyz status", rslt);

            if (rslt == BMA4_OK)
            {
                /* Converting lsb to meter per second squared for 16 bit resolution at 2G range */
                x = lsb_to_ms2(sens_data.x, 2, bma.resolution);
                y = lsb_to_ms2(sens_data.y, 2, bma.resolution);
                z = lsb_to_ms2(sens_data.z, 2, bma.resolution);

                /* Print the data in m/s2 */
                Serial.printf("\n%4.2f, %4.2f, %4.2f", x, y, z);
            }

            /* Decrement the count that determines the number of samples to be printed */
            n_data--;

            /* When the count reaches 0, break and exit the loop */
            if (n_data == 0)
            {
                break;
            }
        }
    }
}

/**
 * @brief Returns the BMA456 Chip ID. This is useful for verification of communication.
 *
 * @return uint8_t Chip ID, value should be 0x16 for BMA456
 */
uint8_t BMA456::getID()
{
    int8_t rslt  = BMA4_OK;
    uint32_t len = 2;
    uint8_t data[len];
    uint8_t add = BMA4_CHIP_ID_ADDR | BMA4_SPI_RD_MASK;

    log_i("Reading accelerometer Chip ID...");
    rslt = spi_read(add, data, len, bma.intf_ptr);
    log_i("Chip ID: 0x%02X", data[1]);
    return rslt;
}

/**
 * @brief Return the temperature of the BMA456 IC.
 * The temperature sensor has 8 bits.
 * The temperature value is defined in Register TEMPERATURE and updated every 1.28 s.
 * The temperature sensor is always on, when the accelerometer sensor is active.
 *  Value   Temperature
 *  0x7F    150 °C
 *  …       …
 * 0x00     23 °C
 * …        …
 * 0x81     -104 °C
 * 0x80 Invalid
 *
 * When there is no valid temperature information available
 * (i.e. last measurement before the time defined above),
 * the temperature indicates an invalid value: 0x80.
 *
 * @return float [°C] Temperature value of IC.
 */
float BMA456::getTemp()
{
    int8_t rslt;
    int32_t tempRaw = 0;
    float temp      = -105;

    rslt = bma4_get_temperature(&tempRaw, BMA4_DEG, &bma);
    if (rslt == BMA4_OK)
    {
        if (((tempRaw - 23) / BMA4_SCALE_TEMP) != 0x80)
        {
            temp = (float)tempRaw / (float)BMA4_SCALE_TEMP;
            log_i("Temperature: %0.0f °C", temp);
        }
        else
            log_i("Temperature reading invalid.");
    }
    return temp;
}

void BMA456::readFIFO()
{
}

void BMA456::tapDetectorTest()
{
    int8_t rslt = BMA4_OK;
    log_i("Tap Detector");

    while (1)
    {
        /* Read interrupt status */
        rslt = bma456h_read_int_status(&int_status, &bma);
        bma4_error_codes_print_result("bma456_read_int_status", rslt);

        if ((rslt == BMA4_OK) & (int_status != 0))
        {
            printf("0x%0X\n", int_status);
            rslt = bma456h_output_state(&tap_out, &bma);

            if (rslt == BMA4_OK)
            {
                /* Enters only if the obtained interrupt is single-tap */
                if (tap_out.single_tap)
                {
                    log_i("Single tap interrupt occurred");
                    printf("Single tap interrupt occurred\n");
                }

                /* Enters only if the obtained interrupt is double-tap */
                else if (tap_out.double_tap)
                {
                    printf("Double tap interrupt occurred\n");
                }

                else if (tap_out.triple_tap)
                {
                    printf("Triple tap interrupt occurred\n");
                }
                tap_iteration--;
            }
        }

        int_status = 0;

        /* Break out of the loop when iteration has reached zero */
        if (tap_iteration == 0)
        {
            printf("Iterations are done. Exiting !");
            tap_iteration = 10;
            break;
        }
    }
}

/**
 * @brief Reads the interrupt status for processing the taps, etc.
 * 
 * @return uint16_t 
 */
uint16_t BMA456::readIntStatus()
{
    int8_t rslt = BMA4_OK;

    rslt = bma456h_read_int_status(&int_status, &bma);
    bma4_error_codes_print_result("bma456_read_int_status", rslt);

    if ((rslt == BMA4_OK) & (int_status != 0))
    {
        // printf("0x%0X\n", int_status);
        rslt = bma456h_output_state(&tap_out, &bma);

        if (rslt == BMA4_OK)
        {
            /* Enters only if the obtained interrupt is single-tap */
            if (tap_out.single_tap)
            {
                // printf("Single tap interrupt occurred\n");
                return 1;
            }

            /* Enters only if the obtained interrupt is double-tap */
            else if (tap_out.double_tap)
            {
                // printf("Double tap interrupt occurred\n");
                return 2;
            }

            else if (tap_out.triple_tap)
            {
                // printf("Triple tap interrupt occurred\n");
                return 3;
            }
        }
    }

    int_status = 0;
    return 0;
}

/*!
 * @brief This internal API converts raw sensor values(LSB) to meters per seconds square.
 */
float BMA456::lsb_to_ms2(int16_t val, float g_range, uint8_t bit_width)
{
    float half_scale = ((float)(1 << bit_width) / 2.0f);

    return (GRAVITY_EARTH * val * g_range) / half_scale;
}

void BMA456::delay_us(uint32_t us, void *intf_ptr)
{
    delayMicroseconds(us);
}

int8_t BMA456::tap_init()
{
    int8_t rslt = BMA4_OK;

    // Enabling features
    rslt |= bma456h_feature_enable(int_en_tap, BMA4_ENABLE, &bma);
    bma4_error_codes_print_result("bma456_feature_enable status", rslt);
    vTaskDelay(3);

    rslt |= bma4_get_int_pin_config(&int_pin_config, int_line_tap, &bma);
    int_pin_config.edge_ctrl = BMA4_LEVEL_TRIGGER;
    int_pin_config.lvl       = BMA4_ACTIVE_HIGH;
    int_pin_config.od        = BMA4_PUSH_PULL;
    int_pin_config.output_en = BMA4_OUTPUT_ENABLE;
    // int_pin_config.input_en  = BMA4_INPUT_DISABLE;
    rslt |= bma4_set_int_pin_config(&int_pin_config, int_line_tap, &bma);
    bma4_error_codes_print_result("bma4_set_int_pin_config status", rslt);
    delay(5);

    // Interrupts
    rslt |= bma456h_map_interrupt(int_line_tap, int_map_tap, BMA4_ENABLE, &bma);
    bma4_error_codes_print_result("bma456_map_interrupt status", rslt);

    /* Getting tap parameters settings */
    rslt = bma456h_tap_get_parameter(&settings, &bma);
    bma4_error_codes_print_result("bma456h_tap_get_parameter status", rslt);
    printf("Setting_8 wait_for_timeout=0x%04X, Setting_2 tap_sens_thres=0x%04X\r\n", settings.wait_for_timeout, settings.tap_sens_thres);

    // settings.tap_sens_thres = 9;
    settings.wait_for_timeout = 1;
    rslt                      = bma456h_tap_set_parameter(&settings, &bma);

    /* Getting tap parameters settings */
    rslt = bma456h_tap_get_parameter(&settings, &bma);
    printf("Setting_8 wait_for_timeout=0x%04X, Setting_2 tap_sens_thres=0x%04X\r\n", settings.wait_for_timeout, settings.tap_sens_thres);

    bma4_set_interrupt_mode(BMA4_NON_LATCH_MODE, &bma);
    vTaskDelay(5);

    // bma456_single_tap_set_sensitivity(0, &bma);
    // bma456_double_tap_set_sensitivity(0, &bma);

    return rslt;
}

/*!
 * @brief Function for initialization of SPI bus.
 */
int8_t BMA456::spi_init(uint8_t clk, uint8_t miso, uint8_t mosi, uint8_t ss)
{
    /* Implement SPI bus initialization according to the target machine. */
    // Accelerometer chip select - active low
    log_i("Initialising SPI for accelerometer");
    _pin_cs = ss;
    pinMode(_pin_cs, OUTPUT);

    // Ensure rising edge to switch to SPI mode.
    digitalWrite(_pin_cs, LOW);  // active
    digitalWrite(_pin_cs, HIGH);  // inactive, rising edge needed to switch to SPI mode

    SPI.begin(clk, miso, mosi, _pin_cs);
    _spiSettings = SPISettings(_spiFrequency, _spiBitOrder, _spiDataMode);

    // Ensure chip is in SPI mode
    getID();
    return 0;
}

/*!
 * @brief Function for reading the sensor's registers through SPI bus.
 */
int8_t BMA456::spi_read(uint8_t reg_addr, uint8_t *reg_data, uint32_t length, void *intf_ptr)
{
    /* Implement the SPI read routine according to the target machine. */
    SPI.beginTransaction(_spiSettings);
    digitalWrite(_pin_cs, LOW);  // assert Slave Select
    // log_d("SPI read - Length %u", length);

    SPI.write(reg_addr | BMA4_SPI_RD_MASK);  // automatically include mask
    SPI.transferBytes(NULL, reg_data, length);

    // for (size_t i = 0; i < length; i++)
    // {
    //     log_d("data[%u]\t0x%02X", i, reg_data[i]);
    // }

    digitalWrite(_pin_cs, HIGH);  // deassert Slave Select
    SPI.endTransaction();
    return BMA4_INTF_RET_SUCCESS;
}

/*!
 * @brief Function for writing the sensor's registers through SPI bus.
 */
int8_t BMA456::spi_write(uint8_t reg_addr, const uint8_t *reg_data, uint32_t length, void *intf_ptr)
{
    /* Implement the SPI write routine according to the target machine. */
    SPI.beginTransaction(_spiSettings);
    digitalWrite(_pin_cs, LOW);  // assert Slave Select
    // log_d("SPI write - Length %u", length);
    SPI.write(reg_addr);
    SPI.writeBytes(reg_data, length);
    digitalWrite(_pin_cs, HIGH);  // deassert Slave Select
    SPI.endTransaction();
    return BMA4_INTF_RET_SUCCESS;
}

/*!
 *  @brief Prints the execution status of the APIs.
 */
void BMA456::bma4_error_codes_print_result(const char api_name[], int8_t rslt)
{
    if (rslt != BMA4_OK)
    {
        log_e("%s\t", api_name);
        if (rslt == BMA4_E_NULL_PTR)
        {
            log_e("Error [%i] : Null pointer", rslt);
        }
        else if (rslt == BMA4_E_CONFIG_STREAM_ERROR)
        {
            log_e("Error [%i] : Invalid configuration stream", rslt);
        }
        else if (rslt == BMA4_E_SELF_TEST_FAIL)
        {
            log_e("Error [%i] : Self test failed", rslt);
        }
        else if (rslt == BMA4_E_INVALID_SENSOR)
        {
            log_e("Error [%i] : Device not found", rslt);
        }
        else
        {
            /* For more error codes refer "*_defs.h" */
            log_e("Error [%i] : Unknown error code", rslt);
        }
    }
}
