#ifndef _BMA456_H  // BMA456_H

#define _BMA456_H

#include <SPI.h>

#include "Arduino.h"
#include "bma456h.h"
#include "bma4_defs.h"

/*! Earth's gravity in m/s^2 */
#define GRAVITY_EARTH (9.80665f)

/*! Macro that holds the total number of accel x,y and z axes sample counts to be printed */
#define ACCEL_SAMPLE_COUNT 1000
class BMA456
{
public:
    BMA456();

    /**
     * @brief Function for initialising the BMA456 Device
     *
     * @return * int8_t
     */
    int8_t init(uint8_t clk, uint8_t miso, uint8_t mosi, uint8_t ss, uint8_t int1 = -1, uint8_t int2 = -1);
    void accTest();
    uint8_t getID();
    float getTemp();
    void readFIFO();
    void tapDetectorTest();
    uint16_t readIntStatus();

private:
    // BMA data structures
    struct bma4_dev bma = {0};                  // Declare an instance of the BMA456 device
    struct bma4_accel_config accel_conf;        // Declare an accelerometer configuration structure
    struct bma4_accel sens_data;                // Declare an instance of the sensor data structure
    struct bma4_int_pin_config int_pin_config;  // Declare an interrupt pin configuration structure

    struct bma456h_out_state tap_out          = {0};
    struct bma456h_multitap_settings settings = {0};

    // test vars
    uint16_t n_data = ACCEL_SAMPLE_COUNT;  // Variable that holds the accelerometer sample count
    float x = 0, y = 0, z = 0;
    uint8_t tap_iteration = 10;

    // SPI Interface ////////////////////
    static SPISettings _spiSettings;
    static uint8_t _pin_cs;
    static const uint8_t _spiBitOrder   = SPI_MSBFIRST;
    static const uint8_t _spiDataMode   = SPI_MODE0;
    static const uint32_t _spiFrequency = 2e6;  // [Hz]
    uint8_t dev_addr                    = 0;    // Variable that holds the SPI chip selection

    // Interrupt interface
    uint8_t _pin_int1;         // Interrupt Pin 1
    uint8_t _pin_int2;         // Interrput Pin 2
    uint16_t int_status  = 0;  // Variable to store accel data ready interrupt status
    uint8_t int_line_tap = BMA4_INTR1_MAP;
    uint16_t int_map_tap = (BMA456H_TAP_OUT_INT);                                                    // Variable to define feature interrupts to be mapped
    uint16_t int_en_tap  = (BMA456H_SINGLE_TAP_EN | BMA456H_DOUBLE_TAP_EN | BMA456H_TRIPLE_TAP_EN);  // Variable to define feature interrupts to be enabled
    // uint16_t int_map_tap  = (BMA456_ANY_MOT_INT );  // Variable to define feature interrupts to be mapped
    uint8_t int_line_fifo = BMA4_INTR2_MAP;
    uint16_t int_map_fifo = (BMA4_FIFO_FULL_INT);  // Variable to define feature interrupts to be mapped

    /**
     * @brief Function for initialising the tap functions
     * 
     */
    int8_t tap_init();

    // Other functions
    static void delay_us(uint32_t us, void *intf_ptr);

    /*!
     *  @brief Function for initialization of SPI bus
     *
     *  @retval 0 -> Success
     *  @retval < 0 -> Failure Info
     */
    int8_t spi_init(uint8_t clk, uint8_t miso, uint8_t mosi, uint8_t ss);

    /*!
     *  @brief Function for reading the sensor's registers through SPI bus.
     *
     *  @param[in] reg_addr     : Register address.
     *  @param[out] reg_data    : Pointer to the data buffer to store the read data.
     *  @param[in] length       : No of bytes to read.
     *  @param[in] intf_ptr     : Interface pointer
     *
     *  @return Status of execution
     *  @retval = BMA4_INTF_RET_SUCCESS -> Success
     *  @retval != BMA4_INTF_RET_SUCCESS  -> Failure Info
     *
     */
    static int8_t spi_read(uint8_t reg_addr, uint8_t *reg_data, uint32_t length, void *intf_ptr);

    /*!
     *  @brief Function for writing the sensor's registers through SPI bus.
     *
     *  @param[in] reg_addr     : Register address.
     *  @param[in] reg_data     : Pointer to the data buffer whose data has to be written.
     *  @param[in] length       : No of bytes to write.
     *  @param[in] intf_ptr     : Interface pointer
     *
     *  @return Status of execution
     *  @retval = BMA4_INTF_RET_SUCCESS -> Success
     *  @retval != BMA4_INTF_RET_SUCCESS  -> Failure Info
     *
     */
    static int8_t spi_write(uint8_t reg_addr, const uint8_t *reg_data, uint32_t length, void *intf_ptr);

    /*!
     *  @brief Prints the execution status of the APIs.
     *
     *  @param[in] api_name : Name of the API whose execution status has to be printed.
     *  @param[in] rslt     : Error code returned by the API whose execution status has to be printed.
     *
     *  @return void.
     */
    void bma4_error_codes_print_result(const char api_name[], int8_t rslt);

    /*!
     *  @brief This internal API converts raw sensor values(LSB) to meters per seconds square.
     *
     *  @param[in] val       : Raw sensor value.
     *  @param[in] g_range   : Accel Range selected (2G, 4G, 8G, 16G).
     *  @param[in] bit_width : Resolution of the sensor.
     *
     *  @return Accel values in meters per second square.
     *
     */
    static float lsb_to_ms2(int16_t val, float g_range, uint8_t bit_width);
};

extern BMA456 bma456;

#endif  // BMA456_H